/*------------------------------------------------------------------------------
  Copyright (c) 2008 Ensolis, LLC. All Rights Reserved.
  ----------------------------------------------------------------------------*/

/******************************************************************************
 * Ping Constants
 *****************************************************************************/
const RUN_INSTALL = Ci.ffIPingService.RUN_INSTALL;
const RUN_UPGRADE = Ci.ffIPingService.RUN_UPGRADE;
const RUN_UNINSTALL = Ci.ffIPingService.RUN_UNINSTALL;

/******************************************************************************
 * Interfaces used monitor upgrade, install, uninstall process and
 * notify the server when changes occur.
 * 
 * @status    FROZEN
 * @version   1.0
 ******************************************************************************/
function PingService() 
{
  //setup additional interfaces
  this._ifaces.push(Ci.nsIObserver);
  this._ifaces.push(Ci.nsISupportsWeakReference);
  
  //setup a new error
  this._error = Cc["@ensolis.com/forecastfox/error-item;1"].
                createInstance(Ci.ffIErrorItem);        
}
PingService.prototype = {
  __proto__: new ServiceBase("PingService"),
  _obSvc: null,
  
  ////////////////////////////
  // nsIObserver 
  observe: function PingService_observe(aSubject, aTopic, aData) 
  {
    //only do something if item is us
    var item = aSubject.QueryInterface(Ci.nsIUpdateItem);
    if (item.id != "{0538E3E3-7E9B-4d49-8831-A227C80A7AD3}")
      return;
    
    //get the migrator service
    var mgrSvc = Cc["@ensolis.com/forecastfox/manager-service;1"].
                 getService(Ci.ffIManagerService);
    var migSvc = mgrSvc.migrator;
        
    //determine action based on message  
    switch (aData) {
        
    //being upgraded
    case "item-upgraded":
        
      //going to version less than tracked
      if (migSvc.compare(item.version, "0.7") < 0)
        this._send(RUN_UNINSTALL, "0.9.7.5", "*");
      break;
          
    //being uninstalled
    case "item-uninstalled":
      this._send(RUN_UNINSTALL, "0.9.7.5", "*"); 
      break;
        
    //being cancelled
    case "item-cancel-action":
      this.run();
      break;
    }
  },

  ////////////////////////////////
  // ffIService  
  
  /**
   * Initialize the component.  Called by the manager service.
   */  
  start: function PingService_start()
  {            
    //get extension manager
    var exMgr = Cc["@mozilla.org/extensions/manager;1"].
                getService(Ci.nsIExtensionManager);
             
    //add observer
    this._obSvc = Cc["@mozilla.org/observer-service;1"].
                  getService(Ci.nsIObserverService);        
    this._obSvc.addObserver(this, "em-action-requested", false);
    
    //return success
    return true;
  },
  
  /**
   * Destroy the component.  Called by the manager service.  This may be
   * called prior to start so it needs to be safe.
   */  
  stop: function PingService_stop()
  {
    //unregister from the observer service
    this._obSvc.removeObserver(this, "em-action-requested");
      
    //remove variables
    this._obSvc = null;
  },
                
  ////////////////////////////////
  // ffIPingService  
    
  /**
   * Run ping command.
   */  
  run: function PingService_run()
  {
    //get preferences
    var pinged = getPref("pinged");
    var migrated = getPref("migrated");
          
    //use migrated value if pinged is blank
    if (pinged == "" && migrated != "")
      pinged = migrated;
    
    //determine ping command and send
    switch (pinged) {
    
      //same version do nothing
      case "0.9.7.5":
        break;
        
      //never installed or previously uninstalled.    
      case "":
      case "*":
        this._send(RUN_INSTALL, "", "0.9.7.5");
        break;
        
      //upgrading from old version
      default:       
        this._send(RUN_UPGRADE, pinged, "0.9.7.5");
        break;
    }
  },
  
  ////////////////////////////
  // Internal Functions 
  
  /**
   * Build query string and send request.
   *
   * @param   Type of ping command to perform.
   * @param   Version we are coming from.
   * @param   Version we are going to.
   */    
  _send: function PingService__send(aType, aFrom, aTo)
  {
    //create query url
    var query = "http://forecastfox.ensolis.com/logs/log.php"; 
    var guid = "{ec8030f7-c20a-464f-9b0e-13a3a9e97384}";
    query += "?type=" + aType;
    query += "&from=" + aFrom;
    query += "&to=";
    query += (aTo == "*") ? "" : aTo;
    query += "&guid=" + guid;
    
    //send ping request    
    
    var request = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"].
                  createInstance(Ci.nsIXMLHttpRequest);      
    request.open("GET", query, true);
    
    //update channel priority - toolkit 1.5 or greater
    if ("nsISupportsPriority" in Ci) {
      if (request.channel instanceof Ci.nsISupportsPriority)
        request.channel.priority = Ci.nsISupportsPriority.PRIORITY_HIGHEST;
    }
    request.send(null);  
    
    //set pref as to version
    setPref("pinged", aTo);
    
    //log the ping in the errors.log for help debugging.  
    var mgrSvc = Cc["@ensolis.com/forecastfox/manager-service;1"].
                 getService(Ci.ffIManagerService);
    var dskSvc = mgrSvc.disk;
    switch (aType) {
    case RUN_INSTALL:
      dskSvc.log("Installing version " + aTo + " on " + guid, null, null);
      break;
    case RUN_UNINSTALL:
      dskSvc.log("Uninstalling version " + aFrom + " on " + guid, null, null);
      break;
    case RUN_UPGRADE:
      dskSvc.log("Upgrading version " + aFrom + " to " + aTo + " on " + guid, 
                 null, null);
      break;
    }
  }
};